/*
	unwings (copyleft) 2004 by Moritz Voss

	 - unpacks and (later) repacks WINGSPAK files for Sldner: Secret Wars
	
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA	
*/


/* 
  This program requires zlib version 1.2.1 or highter
  zlib is Copyright (C) 1995-2003 Jean-loup Gailly and Mark Adler
*/

#define  ZLIB_WINAPI 

#include <fstream>
#include <iostream>

#include <zlib.h>   //good old zlib
#include <direct.h> //for directory manipulation

using namespace std;

#define WINGSPAK_HEADER_LENGTH 12


#define OPTION_OUTPUT_SILENT		0
#define OPTION_OUTPUT_NORMAL		1
#define OPTION_OUTPUT_VERBOSE		2

#define OPTION_MODE_LIST	0
#define OPTION_MODE_UNPACK	1
#define OPTION_MODE_PACK	2


struct tOptions
{ 
	unsigned int output;	//0 silent, 1 normal, 2 verbose
	unsigned int mode;		//0 LIST, 1 unpack, 2 pack
	
	bool force;	
};



std::ifstream pak; // the wingspak file
tOptions options;



class cWPItem
{
public:
	unsigned long run_length;
	unsigned long file_offset;
	unsigned long file_size;
	unsigned char is_directory;
	unsigned char string_length;
	char* name_string;

	cWPItem(void) //instantiate and load info about a new item from the global stream
	{
		string_length = pak.get();
		name_string = new char[string_length+1]; memset(name_string, 0, string_length+1);
		pak.read(name_string, string_length);
			
		if (is_directory = pak.get())
		{
			file_offset = 0;
			file_size	= 0;
		}
		else
		{
			pak.seekg(4, fstream::cur);
			file_offset = (pak.get() << 24) + (pak.get() << 16) + (pak.get() << 8) + pak.get(); //Big Endian, of all things :(
			pak.seekg(4, fstream::cur);
			file_size = (pak.get() << 24) + (pak.get() << 16) + (pak.get() << 8) + pak.get(); //Big Endian, of all things :(
		}			

		run_length    = (pak.get() << 24) + (pak.get() << 16) + (pak.get() << 8) + pak.get(); //Big Endian, of all things :(
	};

	~cWPItem(void)
	{
		delete name_string;
	};

	void store() //uncompress a file using zlib and store to disk, reading from global stream
	{
		std::ofstream outfile;

		void* data = new char[file_size];

		unsigned long oldpos = pak.tellg(); 
		pak.seekg(file_offset);

		//EXCEPTION: This is a LITTLE ENDIAN representation of the size, argh.
		unsigned long dest_size; 
		pak.read((char*) & dest_size, 4);

		void* dest = new char[dest_size]; //dest size now works


		if (file_size)
		{
			pak.read((char*)data, file_size - 4);

			if (int zerror = uncompress((Bytef*)dest, &dest_size, (Bytef*) data, file_size)) 
				if (options.output) cout << "Error: zlib uncompress() failed with error code " << zerror << ".\n";
		} else dest_size = 0; 

		pak.seekg(oldpos);

		if (options.force)
		{
 			outfile.open(name_string, ios_base::out|ios_base::binary|ios_base::trunc);
			outfile.write((char*)dest, dest_size);
			outfile.close(); 
		} else
		{
			std::ifstream testfile;
			testfile.open(name_string); //check with a tentative input open

			if (!testfile) //doesn't exist
			{
				outfile.open(name_string, ios_base::out|ios_base::binary|ios_base::trunc);
				outfile.write((char*)dest, dest_size);
				outfile.close(); 
			} else
			{
				testfile.close();
				char input;
				cout << name_string << " already exists. Overwrite? (y/n/a):";
				cin >> input;
			//	cout << "\n";

				if (input == 'y' || input == 'a')
				{
					outfile.write((char*)dest, dest_size);
					outfile.close(); 
					if (input == 'a') options.force = true;
				}
			}
		}
		delete dest;
		delete data;
	};

};


//retrieve a chunk of data from the pak - recursive, self-terminating
void unPakChunk(unsigned long chunks)
{
	for (unsigned long i=0; (i < chunks) && !pak.eof(); i++) //number of chunks queued
	{
		cWPItem* item = new cWPItem;
		if (item->is_directory)
		{
			switch (options.output)
			{
				case OPTION_OUTPUT_SILENT:  break;
				case OPTION_OUTPUT_VERBOSE:
					cout << "DIRECTORY : " << item->name_string << " : " << item->run_length <<" entries\n";
					break;
				default:
					cout << "D " << item->name_string << "\n";
			};

			if (options.mode == OPTION_MODE_UNPACK)
			{	
				mkdir(item->name_string);
				chdir(item->name_string);
			}
			
			if (item->run_length) unPakChunk(item->run_length);
			
			if (options.mode == OPTION_MODE_UNPACK) chdir("..");
		} else 
		{
			if (options.mode == OPTION_MODE_UNPACK) item->store(); //uncompress and write to disk

			switch (options.output)
			{
				case OPTION_OUTPUT_SILENT:  break;
				case OPTION_OUTPUT_VERBOSE:
					cout << "FILE : " << item->name_string << " : " << item->file_size << " compressed bytes\n";
					break;
				default:
					cout << "F " << item->name_string << "\n";
			};

			unPakChunk(item->run_length-1); //not yet happy with this :( why -1?
		}
	}
};


void unPak(const char* filename, const char* destination)
{
	unsigned long table_offset; //address in pakfile of the table of contents

	pak.open(filename, ios_base::in|ios_base::binary);

	if (!pak)
	{
		if (options.output) cout << "Error: File not found.\n";
	}
	else
	{
		if (chdir(destination)) // not 0 means error
		{
			if (options.output) cout << "Error: target directory does not exist.\n";
		} 
		else
		{
			char header[16];
			pak.read(header, WINGSPAK_HEADER_LENGTH);
			if (!strcmp(header, "WINGSPAK"))
			{
				table_offset = (pak.get() << 24) + (pak.get() << 16) + (pak.get() << 8) + pak.get(); //Big Endian, of all things :(
				pak.seekg(table_offset);
	
				unsigned long root_chunks = (pak.get() << 24) + (pak.get() << 16) + (pak.get() << 8) + pak.get(); //Big Endian, of all things :(

				//unpak the file, using the number of chunks found at the beginning of the table for the root dir
				unPakChunk(root_chunks); 
			}
			else if (options.output) cout << "Error: File is not in WINGSPAK format.\n";
		}
	}
}

//diaply a little help screen
void doHelp(void)
{
	cout << "unwings 0.1.0.0 \n"
			 << "  A tool that uncompresses WINGSPAK files using zlib.\n"
			 << "  (c)2004 by Moritz Voss (public@thygrrr.de)\n"
			 << "\n"
			 << "  This program is free software; you can redistribute it and/or\n"
			 << "  modify it under the terms of the GNU General Public License.\n"
			 << "\n"
			 << "Usage:\n"
			 << "  unwings -<options> <pak-file> [target-dir]\n"
			 << "\n"
			 << "Options:\n"
			 << "  x : extract contents (default: into current directory)\n"
			 << "  l : list contents\n"
			 << "  f : assume 'yes' on all questions (e.g. overwrite file?)\n"
			 << "  s : suppress all output except interactive warnings\n"
			 << "  n : normal output (default)\n"
			 << "  v : verbose output\n"
			 << "\n"
			 << "Examples:\n"
			 << "   unwings -x soldner1.pak mystuff \n"
			 << "    --> unpacks soldner1.pak into directory 'mystuff' \n"
			 << "   unwings -lv soldner2.pak\n"
			 << "    --> outputs verbose content list for soldner2.pak \n"
			 << "   unwings -xf soldner1.pak\n"
			 << "    --> unpacks soldner1.pak into current directory and \n"
			 << "        overwrites all existing files without warning\n";

	/*
			 << "\n"
			 << "\n"
			 <<	"Compress a file \n"
			 << "    unwings -c <pak-file> <path> \n"
			 << "\n"
			 << "NOTICE : creation of pak files not yet implemented\n";		
	*/
	exit(0); //shutdown program
}


// main function, just parse the command line parameters, then call actual worker functions
int main( int argc, char *argv[ ] )
{
	switch (argc)
	{
		case 3: 
		case 4:
			{
				if (*argv[1] == '-') //it's a valid options string
				{
					options.output = OPTION_OUTPUT_NORMAL; //default setting here

					if (strchr(argv[1], 'x')) //user wants to unpack
					{
						if (strchr(argv[1], 'c') || strchr(argv[1], 'l')) doHelp(); //c, l and x are mutually exclusive
						options.mode = OPTION_MODE_UNPACK;
					}

					if (strstr(argv[1], "c")) //user wants to pack
					{
						if (strchr(argv[1], 'x') || strchr(argv[1], 'l')) doHelp(); //c, l and x are mutually exclusive
						options.mode = OPTION_MODE_PACK;
					}

					if (strstr(argv[1], "l")) //user wants to list
					{
						if (strchr(argv[1], 'x') || strchr(argv[1], 'c')) doHelp(); //c, l and x are mutually exclusive
						options.mode = OPTION_MODE_LIST;
					}

					if (strchr(argv[1], 's')) //user wants to list
					{
						if (strchr(argv[1], 'v')) doHelp(); //v and s are mutually exclusive
						options.output = OPTION_OUTPUT_SILENT;
					}
					
					if (strchr(argv[1], 'v')) //user wants to list
					{
						if (strchr(argv[1], 's')) doHelp(); //v and s are mutually exclusive
						options.output = OPTION_OUTPUT_VERBOSE;
					}

					options.force = (strchr(argv[1], 'f') != NULL); //force overwrite of existing files
					switch (options.mode)
					{
						case OPTION_MODE_UNPACK:
						{
							if (argc == 4) unPak(argv[2], argv[3]);
								else unPak(argv[2], ".");
							break;
						}
						case OPTION_MODE_PACK:
						{
							doHelp();
							break;
						}
						default:
							unPak(argv[2], ".");							
					}
				}
				else doHelp();
			}
			break;

		default: 
			doHelp();
	};

	return 0;
}
